(* ----------------------------------------------------------------- *)
(* Turbo-Pascal Unit, ab TP 6

   Unterroutinen fr Datentransfer (ASCII oder binr) ber eine
   serielle Schnittstelle. Sowohl der Datenempfang als auch das
   Senden von Daten werden interrupt-gepuffert.
   Nach dem Programmstart mssen mit ComSet die von der Hardware
   bereitgestellte Port- und Interruptadressen eingestellt werden:
   z.B.:
     Standardeinstellung fr COM1: 3F8 mit IRQ 4
     Standardeinstellung fr COM2: 2F8 mit IRQ 3
   Die Portadressen knnen dem Array "ComPort[ComNr]" entnommen werden
   (ComNr = 1..4).
   Vor Programmende sollten mit ComRls die Interruptvektoren wieder auf
   ihren Urzustand zurckgesetzt werden.

   Status: (ComInit und ComStat)
           Bit 15 = 0
               14 = Transmitter shift reg. empty
               13 = Transmitter holding register empty
               12 = Break interrupt
               11 = Framing error
               10 = Parity error
                9 = Overrun error
                8 = Data received (buffer not empty)
                7 = Received line signal detect (RLSD)
                6 = Ring indicator (RI)
                5 = Data set ready (DSR)
                4 = Clear to send (CTS)
                3 = Delta received lin signal detect
                2 = Trailing edge ring indicator
                1 = Delta data set ready
                0 = Delta clear to send

   J. Rathlev, Mai  1993
       Rev. 1, Jan. 1998
       Rev. 2; Sep. 2000   mit XOn/XOff (siehe Var. UseXoff) *)

unit TPComNX;

(* ======================================================================= *)
interface

uses Dos, Crt;

const
  ComPortMax = 4;
  ComPort: array [1..ComPortMax] of word = ($3F8,$2F8,$338,$238);
  ComIRQ : array [1..ComPortMax] of byte = (4,3,9,5);

  NoPar = 'N';
  EvenPar = 'E';
  OddPar = 'O';

  ParityModes = [Nopar,EvenPar,OddPar];
  Parities : array [0..2] of char = (Nopar,EvenPar,OddPar);

  UseXOff : boolean = true;

  Komma = ',';
  XOn = ^Q;
  XOff = ^S;

  (* nachfolgende Variable kann vom Hauptprogramm verndert werden *)
  ComOutWait : boolean = false;
  (* false: bei Ausgabe ber COM nicht warten bis Schnittstelle bereit
            (Zeichen gehen verloren)
     true : bei Ausgabe warten bis Schnittstelle bereit *)

(* ----------------------------------------------------------------------- *)
(* Standard-Com initialisieren *)
function InitCom (Nr : word) : boolean;

(* ----------------------------------------------------------------------- *)
(* Interruptvektor setzen *)
procedure ComSet (CPort : word;
                  IntAd : byte);

(* ----------------------------------------------------------------------- *)
(* Interruptvektor zurcksetzen *)
procedure ComRls;

(* ----------------------------------------------------------------------- *)
(* serielle Schnittstelle initialisieren und Modem einschalten *)
procedure ComInit (Baud        : longint;
                   Par         : char;
                   Len,Stb     : word;
               var Status      : word);
(* Baud  = Baud Rate (110 ... 115200)
   Par   = Paritt (N,E,O)
   Len   = Anzahl der Datenbits (5,6,7,8)
   Stb   = Anzahl der Stopbits (1,2)
   Status = Line und Modem Status *)

(* ----------------------------------------------------------------------- *)
(* ser. Schnittstelle CNr mit Mode-String CPar initialisieren (siehe auch
   DOS-Befehl "MODE" *)
procedure SetComMode (var ComPar : String;
                      var Err    : boolean;
                      var Status : word);

(* ----------------------------------------------------------------------- *)
(* Schnittstellenbuffer lschen und Modem einschalten *)
procedure ComOn;

(* ----------------------------------------------------------------------- *)
(* Modem ausschalten *)
procedure ComOff;

(* ----------------------------------------------------------------------- *)
(* Status der ser. Schnittstelle abfragen *)
function ComStat : word;

(* ----------------------------------------------------------------------- *)
(* LCR-Status der ser. Schnittstelle abfragen *)
function ComLCR : byte;

(* ----------------------------------------------------------------------- *)
(* Empfangsbuffer Count der ser. Schnittstelle abfragen *)
function ComRcvBuf : word;

(* ----------------------------------------------------------------------- *)
(* Zeilenzhler der ser. Schnittstelle abfragen *)
function ComRcvLines : word;

(* ----------------------------------------------------------------------- *)
(* Sendebuffer Count der ser. Schnittstelle abfragen *)
function ComSndBuf : word;

(* prfen, ob Sendebuffer voll ist*)
function ComSndBufNotFull : boolean;

(* ----------------------------------------------------------------------- *)
(* Warten auf Input ber die ser. Schnittstelle *)
function ComRec (TimeOut : integer) : boolean;
(* Timeout = Zeit in 1/10 Sek., die gewartet wird. Wird in
             dieser Zeit ein Zeichen empfangen, ist ComRec=true *)

(* ----------------------------------------------------------------------- *)
(* Zeichen von ser. Schnittstelle einlesen *)
function ComRead (var ErrStat : byte) : char;
(* ErrStat = Fehlerstatus : Bit 1 = Overrun Error
                            Bit 2 = Parity Error
                            Bit 3 = Framing Error
                            Bit 4 = Break Interrupt *)

(* ----------------------------------------------------------------------- *)
(* Zeichen an ser. Schnittstelle ausgeben *)
procedure ComWrite (Zeichen : char);
(* Zeichen = auszugebendes Zeichen *)

(* ----------------------------------------------------------------------- *)
(* String von ser. Schnittstelle einlesen *)
function ComRdStr (var ErrStat : byte) : string;
(* ErrStat = Fehlerstatus : Bit 1 = Overrun Error
                            Bit 2 = Parity Error
                            Bit 3 = Framing Error
                            Bit 4 = Break Interrupt
   ComRdStr = String bis Carriage Return *)

(* ----------------------------------------------------------------------- *)
(* String an ser. Schnittstelle ausgeben *)
procedure ComWrStr (Zeile : string);
(* Zeile = zu schreibender String (ohne CR/LF) *)

(* ----------------------------------------------------------------------- *)
(* CR/LF an ser. Schnittstelle ausgeben *)
procedure ComWrCrLf;

(* ----------------------------------------------------------------------- *)
(* Integer-Zahl an ser. Schnittstelle ausgeben *)
procedure ComWrInt (x,n   : integer);
(* x     = Integerzahl
   n     = Anzahl der Stellen *)

(* ----------------------------------------------------------------------- *)
(*$IFOPT N+ *)
(* Real-Zahl an ser. Schnittstelle ausgeben *)
procedure ComWrReal (x     : double;
                     n,d   : integer);
(* x     = Realzahl
   n     = Anzahl der Stellen
   d     = Anzahl der Stellen hinter dem Dezimalpunkt *)
(*$ENDIF *)

(* ======================================================================= *)
implementation
const
  IntCtlAdr = $20;       { Portadresse des Int.-Contr. }
  IntMskAdr = $21;       { Portadresse des Int.Maskreg. }
  BufL    =   1024;      { Buffer Lnge }

  { Register des 8250 }
  IER     =     1;       {  Interrupt Enable Register }
  IIR     =     2;       {  Interrupt Identification Register }
  LCR     =     3;       {  Line Control Register }
  MCR     =     4;       {  MODEM Control Register }
  LSR     =     5;       {  Line Status Register }
  MSR     =     6;       {  MODEM Status Register }

  PortAdr : word = 0;    { RS232 port Adr. }

type
  BufTyp = record
    Ch  : char;
    Err : byte;
    end;

var
  RBuffer  : array [0..BufL-1] of BufTyp;  {input buffer }
  WBuffer  : array [0..BufL-1] of char;    {output buffer }
  RBCnt,                   { read buffer count }
  RBLines,                 { read line count }
  RBiPnt,                  { read input buffer pointer }
  RBoPnt,                  { read output buffer pointer }
  WBCnt,                   { write buffer count }
  WBiPnt,                  { write input buffer pointer }
  wBoPnt,                  { write  output buffer pointer }
  HwInt       : word;      { IRQ Number }
  HwiPtr      : pointer;   { Pointer to old Hardware Int. }
  OldExitProc : Pointer;   { Saves exit procedure address }
  Pause       : boolean;   { true after XOff, false after XOn }

const
  HwMsk       : byte = 0; { IRQ Mask }

(* -------------------------------------------------------------------------
   Interrupt Routine *)
procedure ComInt; interrupt;
var
  c     : char;
  ist   : byte;
  ende  : boolean;
begin
  asm sti end;
  repeat
    ist:=Port[PortAdr+IIR];
    ende:=ist and 1=1;
    if not ende then begin
      ist:=ist and $6;
      if ist=4 then begin  (* empfangen *)
        Port[PortAdr+MCR]:=Port[PortAdr+MCR] and $FE;  { DTR off }
        c:=chr(Port[PortAdr]);
        if UseXOff and (c=XOff) then Pause:=true
        else if UseXOff and (c=XOn) then begin (* Starte Senden *)
          Pause:=false;
          Port[PortAdr+IER]:=Port[PortAdr+IER] or 2;   { enable send interrupt }
          end
        else with RBuffer[RBiPnt] do begin
          Err:=Port[PortAdr+LSR] and $1E;
          Ch:=c;
          if Ch=#13 then inc(RBLines);
          end;
        if RBCnt<BufL then begin
          inc(RBCnt); inc(RBiPnt);
          if RBiPnt=BufL then RBiPnt:=0;
          end;
        Port[PortAdr+MCR]:=Port[PortAdr+MCR] or 1;  { DTR on }
        end;
      if ist=2 then begin (* senden *)
        if (WBCnt=0) or Pause then begin  (* stopp *)
          Port[PortAdr+IER]:=Port[PortAdr+IER] and $D;
          end
        else if not pause then begin
          dec(WBCnt);
          Port[PortAdr]:=ord(WBuffer[WBoPnt]);
          inc(WBoPnt);
          if WBoPnt=BufL then WBoPnt:=0;
          end;
        end;
      if ist=0 then ist:=Port[PortAdr+MSR];
      if ist=6 then ist:=Port[PortAdr+LSR];
      end;
    until ende;
  asm cli end;
  Port[IntCtlAdr]:=$60+HwInt;
  end;

(* ----------------------------------------------------------------------- *)
(* Schnittstellenbuffer lschen und Modem einschalten (identisch mit ComClr *)
procedure ComOn;
var
  b  : byte;
begin
  if PortAdr>0 then begin
    asm cli; end;
    RBCnt:=0; RBLines:=0; RBoPnt:=0; RBiPnt:=0;
    WBCnt:=0; WBoPnt:=0; WBiPnt:=0;
    asm sti; end;
    b:=Port[PortAdr];         { clear receiver buffer reg. }
    Port[PortAdr+MCR]:=$0B;   { set DTR, RTS and OUT2 }
    Port[PortAdr+IER]:=1;     { enable interrupt for receive  }
    Port[IntMskAdr]:=Port[IntMskAdr] and (not HwMsk); { set interrupt mask }
    Pause:=false;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Modem ausschalten *)
procedure ComOff;
begin
  if PortAdr>0 then begin
    Port[IntMskAdr]:=Port[IntMskAdr] or HwMsk;  { reset interrupt mask }
    Port[PortAdr+IER]:=0;         { IER}
    Port[PortAdr+MCR]:=0;         { set modem off }
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Status der ser. Schnittstelle abfragen *)
function ComStat : word;
(* Status: Bit 15 = 0
               14 = Transmitter shiftv reg. empty
               13 = Transmitter holding register empty
               12 = Break interrupt
               11 = Framing error
               10 = Parity error
                9 = Overrun error
                8 = Data received (buffer not empty)
                7 = Received line signal detect (RLSD)
                6 = Ring indicator (RI)
                5 = Data set ready (DSR)
                4 = Clear to send (CTS)
                3 = Delta received lin signal detect
                2 = Trailing edge ring indicator
                1 = Delta data set ready
                0 = Delta clear to send *)
var
  BStat : word;
begin
  if PortAdr>0 then begin
    if RBCnt=0 then BStat:=0 else BStat:=$100;
    ComStat:=256*(Port[PortAdr+LSR] and $7E)+Port[PortAdr+MSR]+BStat;
    end
  else ComStat:=0;
  end;

(* ----------------------------------------------------------------------- *)
(* Line Control Register der ser. Schnittstelle abfragen
   siehe Datenblatt 8250 *)
function ComLCR : byte;
begin
  if PortAdr>0 then ComLCR:=Port[PortAdr+LCR]
  else ComLCR:=0;
  end;

(* -------------------------------------------------------------------------
(* Status prfen *)
function StatusOK (Rg,Mask : byte;
                   TOCnt   : word) : boolean;
(* Rg   = Modem Register
   Mask = zu testende Bits
   TOCnt = Wartezeit in ms
   StatusOK = false : Time Out *)
var
  st   : byte;
  TCnt : word;
begin
  if PortAdr>0 then begin
    TCnt:=0;
    while (Port[PortAdr+Rg] and Mask<>Mask) and (TCnt<TOCnt) do begin
      delay(1); inc(TCnt);
      end;
    StatusOK:=TCnt<TOCnt;
    end
  else StatusOK:=false;
  end;

(* -------------------------------------------------------------------------
(* serielle Schnittstelle initialisieren und Modem einschalten*)
procedure ComInit (Baud        : longint;
                   Par         : char;
                   Len,Stb     : word;
               var Status      : word);
(* Baud  = Baud Rate (110 ... 115200)
   Par   = Paritt (N,E,O)
   Len   = Anzahl der Datenbits (5,6,7,8)
   Stb   = Anzahl der Stopbits (1,2)
   Status = Line und Modem Status (siehe auch ComStat) *)
var
   reg : registers;
   i,k :integer;
begin
  if PortAdr>0 then begin
    ComOff;
    i:=Baud div 10;
    case i of    (* Baud Rate *)
          5 : k:=2304;
          7 : k:=1536;
         11 : k:=1047;
         13 : k:=857;
         15 : k:=768;
         30 : k:=384;
         60 : k:=192;
        120 : k:=96;
        180 : k:=64;
        240 : k:=48;
        480 : k:=24;
        960 : k:=12;
       1920 : k:=6;
       3840 : k:=3;
       5760 : k:=2;
      11520 : k:=1;
         else k:=48; (* 2400 baud *)
         end;
    Port[PortAdr+LCR]:=$80;       { set DLAB }
    Port[PortAdr]:=lo(k);
    Port[PortAdr+1]:=hi(k);
    case Par of     (* Paritt *)
      'n','N': i:=0;
      'o','O': i:=1;
      'e','E': i:=3;
      else i:=0;
      end;
    k:=8*i;
    case Stb of     (* Stop Bits *)
      1 : i:=0;
      2 : i:=1;
      else i:=0;
      end;
    k:=k+4*i;
    case Len of     (* Anzahl der Bits *)
      5 : i:=0;
      6 : i:=1;
      7 : i:=2;
      8 : i:=3;
      else i:=3;
      end;
    k:=k+i;
    Port[PortAdr+LCR]:=k;       { set LCR }
    ComOn;
    Status:=ComStat;
    end;
  end;

{ ------------------------------------------------------------------- }
(* Integer, Double oder String aus einem String s bis zum nchsten
   Komma lesen
   s wird um den verarbeiteten Teil gekrzt *)
function ReadNxtInt (var s   : OpenString;
                     Del     : char;
                     var err : boolean) : integer;
var
  sl     : byte absolute s;
  n,i,ic : integer;
begin
  i:=pos (Del,s);
  if i=0 then i:=succ(sl);
  val(copy(s,1,pred(i)),n,ic);
  if ic=0 then ReadNxtInt:=n
  else begin
    ReadNxtInt:=0;
    err:=true;
    end;
  delete(s,1,i);
  end;

{ ------------------------------------------------------------------- }
function ReadNxtStr (var s   : OpenString;
                     Del     : char) : string;
var
  sl   : byte absolute s;
  i,ic : integer;
begin
  i:=pos (Del,s);
  if i=0 then i:=succ(sl);
  ReadNxtStr:=copy(s,1,pred(i));
  delete(s,1,i);
  end;

(* ----------------------------------------------------------------------- *)
(* ser. Schnittstelle CNr mit Mode-String CPar initialisieren (siehe auch
   DOS-Befehl "MODE" *)
procedure SetComMode (var ComPar : String;
                      var Err    : boolean;
                      var Status : word);
var
  baud,nb,sb : integer;
  par        : string[1];
  c          : char;
begin
  baud:=ReadNxtInt (ComPar,Komma,err);
  par:=ReadNxtStr (ComPar,Komma);
  c:=UpCase(par[1]);
  if c in ParityModes then Par:=c
  else err:=true;
  nb:=ReadNxtInt (ComPar,Komma,err);
  sb:=ReadNxtInt (ComPar,Komma,err);
  if not err then ComInit (baud,par[1],nb,sb,status);
  end;

(* ----------------------------------------------------------------------- *)
(* Empfangsbuffer Count der ser. Schnittstelle abfragen *)
function ComRcvBuf : word;
begin
   ComRcvBuf:=RBCnt;
   end;

(* ----------------------------------------------------------------------- *)
(* Zeilenzhler der ser. Schnittstelle abfragen *)
function ComRcvLines : word;
begin
   ComRcvLines:=RBLines;
   end;

(* ----------------------------------------------------------------------- *)
(* Sendebuffer Count der ser. Schnittstelle abfragen *)
function ComSndBuf : word;
begin
   ComSndBuf:=WBCnt;
   end;

(* ----------------------------------------------------------------------- *)
(* prfen, ob Sendebuffer voll ist*)
function ComSndBufNotFull : boolean;
begin
   ComSndBufNotFull:=WBCnt<BufL;
   end;

(* ----------------------------------------------------------------------- *)
(* Warten auf Input ber die ser. Schnittstelle *)
function ComRec (TimeOut : integer) : boolean;
(* Timeout = Zeit in 1/10 Sek., die gewartet wird. Wird in
             dieser Zeit ein Zeichen empfangen, ist ComRec=true *)
begin
   while ((ComStat and $100)=0) and (TimeOut>=0) do begin
     dec(TimeOut); delay (100);
     end;
   if TimeOut<0 then ComRec:=false else ComRec:=true;
   end;

(* ----------------------------------------------------------------------- *)
(* Zeichen von ser. Schnittstelle einlesen *)
function ComRead (var ErrStat : byte) : char;
(* ErrStat = Fehlerstatus : Bit 1 = Overrun Error
                            Bit 2 = Parity Error
                            Bit 3 = Framing Error
                            Bit 4 = Break Interrupt
                            Bit 7 = Time Out *)
const
  TOCnt = 10;
var
  TCnt : word;
begin
  TCnt:=0; ComRead:=#0;
  while (RBCnt=0) and (TCnt<TOCnt) do begin
    delay (1); inc(TCnt);
    end;
  if RBCnt>0 then begin
    dec(RBCnt);
    with RBuffer[RBoPnt] do begin
      ComRead:=Ch; ErrStat:=Err;
      if (Ch=#13) and (RBLines>0) then dec(RBLines);
      end;
    inc(RBoPnt); if RBoPnt=BufL then RBoPnt:=0;
    end
  else ErrStat:=$80;
  end;

(* ----------------------------------------------------------------------- *)
(* Zeichen an ser. Schnittstelle ausgeben *)
procedure ComWrite (Zeichen : char);
(* Zeichen = auszugebendes Zeichen *)
var
  err : boolean;
begin
  repeat
    err:=WBCnt>=BufL;
    if not err then begin
      asm cli; end;
      if (WBCnt=0) and not Pause then begin  (* 1. Zeichen *)
        Port[PortAdr+IER]:=Port[PortAdr+IER] or 2;   { enable send interrupt }
        end;
      inc(WBCnt);
      WBuffer[WBiPnt]:=Zeichen;
      inc(WBiPnt); if WBiPnt=BufL then WBiPnt:=0;
      asm sti; end;
      end;
    err:=ComOutWait and err;
    if err then delay (100);
    until not err;
  end;

(* ----------------------------------------------------------------------- *)
(* String von ser. Schnittstelle einlesen mit Ctrl-Q/S-Prokoll
   Achtung: Die Fehlerbehandlung mu im aufrufrenden Programm erfolgen:
     repeat s:=ComRdStr(Err) until Err=0;  *)
function ComRdStr (var ErrStat : byte) : string;
(* ErrStat = Fehlerstatus : Bit 1 = Overrun Error
                            Bit 2 = Parity Error
                            Bit 3 = Framing Error
                            Bit 4 = Break Interrupt
                            Bit 7 = Time Out
   ComRdStr = String bis Carriage Return *)
var
  z : char;
  s : string;
begin
  s:='';
  repeat
    z:=ComRead(ErrStat);
    if z>#31 then s:=s+z;
    until (z=#13) or not ComRec(10) {or (ErrStat<>0)};
  ComRdStr:=s;
  end;

(* ----------------------------------------------------------------------- *)
(* String an ser. Schnittstelle ausgeben *)
procedure ComWrStr (Zeile : string);
(* Zeile = zu schreibender String (ohne CR/LF) *)
var
  i : integer;
  l : byte absolute Zeile;
begin
  for i:=1 to l do ComWrite (Zeile[i]);
  end;

(* ----------------------------------------------------------------------- *)
(* CR/LF an ser. Schnittstelle ausgeben *)
procedure ComWrCrLf;
begin
  ComWrite (#13); ComWrite (#10);
  end;

(* ----------------------------------------------------------------------- *)
(* Integer-Zahl an ser. Schnittstelle ausgeben *)
procedure ComWrInt (x,n   : integer);
(* x     = Integerzahl
   n     = Anzahl der Stellen *)
var
  number : string [10];
begin
  str (x:n,number); ComWrStr (number);
  end;

(* ----------------------------------------------------------------------- *)
(*$IFOPT N+ *)
(* Real-Zahl an ser. Schnittstelle ausgeben *)
procedure ComWrReal (x     : double;
                     n,d   : integer);
(* x     = Realzahl
   n     = Anzahl der Stellen
   d     = Anzahl der Stellen hinter dem Dezimalpunkt *)
var
  number : string [20];
begin
  str (x:n:d,number); ComWrStr (number);
  end;
(*$ENDIF *)

(* ----------------------------------------------------------------------- *)
(* Standard-Com initialisieren *)
function InitCom (Nr : word) : boolean;
begin
  if Nr<=ComPortMax then begin
    Comset (ComPort[Nr],ComIrq[Nr]);
    InitCom:=true;
    end
  else InitCom:=false;
  end;

(* ----------------------------------------------------------------------- *)
(* Interruptvektor setzen *)
procedure ComSet (CPort : word;
                  IntAd : byte);
begin
  GetIntVec (IntAd+8,HwIPtr);
  SetIntVec (IntAd+8,@ComInt);
  HwMsk:=1 shl IntAd; HwInt:=IntAd;
  PortAdr:=CPort;
  end;

(* ----------------------------------------------------------------------- *)
(* Interruptvektor zurcksetzen *)
procedure ComRls;
begin
  if PortAdr>0 then begin
    ComOff;
    SetIntVec (HwInt+8,HwIPtr);
    PortAdr:=0;
    end;
  end;

(* ----------------------------------------------------------------------- *)
{$F+}
(* Interruptvektor bei Abbruch zurcksetzen *)
procedure ComExitProc;
begin
  ExitProc := OldExitProc; { Restore exit procedure address }
  ComRls;
  end;
{$F-}

(* ----------------------------------------------------------------------- *)
begin
  OldExitProc := ExitProc;                { save previous exit proc }
  ExitProc := @ComExitProc;               { insert our exit proc in chain }
  RBCnt:=0; RBLines:=0; RBoPnt:=0; RBiPnt:=0;
  WBCnt:=0; WBoPnt:=0; WBiPnt:=0;
  end.
